;; Generated package description from devil.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "devil" "0.7.3.0.20240129.2809" "Minor mode for translating key sequences" '((emacs "24.4")) :commit "c2dc9fc2886471daff654e12b46a144a4d1361ae" :authors '(("Susam Pal" . "susam@susam.net")) :maintainer '("Susam Pal" . "susam@susam.net") :keywords '("convenience" "abbrev") :url "https://github.com/susam/devil")
